var __defProp = Object.defineProperty;
var __name = (target, value) => __defProp(target, "name", { value, configurable: true });

// src/index.ts
import { Context, Service, z } from "cordis";
import { Binary, defineProperty, isNullable } from "cosmokit";
import { createRequire } from "node:module";
import fetchFile from "@cordisjs/fetch-file";
var kHTTPError = Symbol.for("cordis.http.error");
var HTTPError = class extends Error {
  constructor(message, code) {
    super(message);
    this.code = code;
  }
  static {
    __name(this, "HTTPError");
  }
  [kHTTPError] = true;
  response;
  static is(error) {
    return !!error?.[kHTTPError];
  }
};
function encodeRequest(data) {
  if (data instanceof URLSearchParams) return [null, data];
  if (data instanceof ArrayBuffer) return [null, data];
  if (ArrayBuffer.isView(data)) return [null, data];
  if (data instanceof Blob) return [null, data];
  if (data instanceof FormData) return [null, data];
  return ["application/json", JSON.stringify(data)];
}
__name(encodeRequest, "encodeRequest");
var validateStatus = /* @__PURE__ */ __name((status) => status < 400, "validateStatus");
var HTTP = class _HTTP extends Service {
  constructor(ctx, config = {}) {
    super(ctx, "http");
    this.config = config;
    this.decoder("json", (raw) => raw.json());
    this.decoder("text", (raw) => raw.text());
    this.decoder("blob", (raw) => raw.blob());
    this.decoder("arraybuffer", (raw) => raw.arrayBuffer());
    this.decoder("formdata", (raw) => raw.formData());
    this.decoder("stream", (raw) => raw.body);
    this.proxy(["http", "https"], (url) => {
      return new this.undici.ProxyAgent(url.href);
    });
    this.ctx.on("http/fetch", async (url, init, config2, next) => {
      if (url.protocol !== "file:") return next();
      if (init.method !== "GET") {
        return new Response(null, { status: 405, statusText: "Method Not Allowed" });
      }
      return fetchFile(url, init, {
        attachment: true,
        onError: ctx.logger?.error
      });
    }, { prepend: true });
    this.ctx.on("http/fetch", async (url, init, config2, next) => {
      const capture = /^data:([^,]*),(.*)$/.exec(url.href);
      if (!capture) return next();
      if (init.method !== "GET") {
        return new Response(null, { status: 405, statusText: "Method Not Allowed" });
      }
      let [, type, data] = capture;
      let bodyInit = data;
      if (type.endsWith(";base64")) {
        type = type.slice(0, -7);
        bodyInit = Binary.fromBase64(data);
      } else {
        bodyInit = decodeURIComponent(data);
      }
      return new Response(bodyInit, {
        status: 200,
        statusText: "OK",
        headers: { "content-type": type }
      });
    }, { prepend: true });
  }
  static {
    __name(this, "HTTP");
  }
  static inject = {
    logger: { required: false }
  };
  static Error = HTTPError;
  static undici;
  static {
    const require2 = createRequire(import.meta.url);
    try {
      if (process.execArgv.includes("--expose-internals")) {
        this.undici = require2("internal/deps/undici/undici");
      } else {
        this.undici = require2("undici");
      }
    } catch {
    }
    for (const method of ["get", "delete"]) {
      defineProperty(_HTTP.prototype, method, async function(url, config) {
        const response = await this(url, { method, validateStatus, ...config });
        return response.data;
      });
    }
    for (const method of ["patch", "post", "put"]) {
      defineProperty(_HTTP.prototype, method, async function(url, data, config) {
        const response = await this(url, { method, data, validateStatus, ...config });
        return response.data;
      });
    }
  }
  static Config = z.object({
    timeout: z.natural().role("ms").description("等待请求的最长时间。"),
    keepAlive: z.boolean().description("是否保持连接。"),
    proxyAgent: z.string().description("代理服务器地址。")
  });
  static Intercept = z.object({
    baseURL: z.string().description("基础 URL。"),
    timeout: z.natural().role("ms").description("等待请求的最长时间。"),
    keepAlive: z.boolean().description("是否保持连接。"),
    proxyAgent: z.string().description("代理服务器地址。")
  });
  isError = HTTPError.is;
  _decoders = /* @__PURE__ */ Object.create(null);
  _proxies = /* @__PURE__ */ Object.create(null);
  get undici() {
    if (_HTTP.undici) return _HTTP.undici;
    throw new Error("please install `undici`");
  }
  static mergeConfig = /* @__PURE__ */ __name((target, source) => ({
    ...target,
    ...source,
    headers: {
      ...target?.headers,
      ...source?.headers
    }
  }), "mergeConfig");
  decoder(type, decoder) {
    return this.ctx.effect(() => {
      this._decoders[type] = decoder;
      return () => delete this._decoders[type];
    });
  }
  proxy(name, factory) {
    return this.ctx.effect(() => {
      for (const key of name) {
        this._proxies[key] = factory;
      }
      return () => {
        for (const key of name) {
          delete this._proxies[key];
        }
      };
    });
  }
  extend(config = {}) {
    return this[Service.extend]({
      config: _HTTP.mergeConfig(this.config, config)
    });
  }
  resolveConfig(init) {
    let result = { headers: {}, ...this.config };
    this.ctx.emit(this, "http/config", result);
    let intercept = this.ctx[Context.intercept];
    while (intercept) {
      result = _HTTP.mergeConfig(result, intercept.http);
      intercept = Object.getPrototypeOf(intercept);
    }
    result = _HTTP.mergeConfig(result, init);
    return result;
  }
  resolveURL(url, config, isWebSocket = false) {
    try {
      url = new URL(url, config.baseURL);
    } catch (error) {
      throw new TypeError(`Invalid URL: ${url}`);
    }
    if (isWebSocket) url.protocol = url.protocol.replace(/^http/, "ws");
    for (const [key, value] of Object.entries(config.params ?? {})) {
      if (isNullable(value)) continue;
      url.searchParams.append(key, value);
    }
    return url;
  }
  defaultDecoder(response) {
    const type = response.headers.get("content-type");
    if (type?.startsWith("application/json")) {
      return response.json();
    } else if (type?.startsWith("text/")) {
      return response.text();
    } else {
      return response.arrayBuffer();
    }
  }
  async [Service.invoke](...args) {
    let method;
    if (typeof args[1] === "string" || args[1] instanceof URL) {
      method = args.shift();
    }
    const config = this.resolveConfig(args[1]);
    const url = this.resolveURL(args[0], config);
    method ??= config.method ?? "GET";
    const controller = new AbortController();
    if (config.signal) {
      if (config.signal.aborted) {
        throw config.signal.reason;
      }
      config.signal.addEventListener("abort", () => {
        controller.abort(config.signal.reason);
      });
    }
    const dispose = this.ctx.effect(() => {
      const timer = config.timeout && setTimeout(() => {
        controller.abort(new HTTPError("request timeout", "ETIMEDOUT"));
      }, config.timeout);
      return () => {
        clearTimeout(timer);
      };
    });
    controller.signal.addEventListener("abort", () => dispose());
    try {
      const headers = new Headers(config.headers);
      const init = {
        method,
        headers,
        body: config.data,
        keepalive: config.keepAlive,
        redirect: config.redirect,
        signal: controller.signal
      };
      if (config.data && typeof config.data === "object") {
        const [type, body] = encodeRequest(config.data);
        init.body = body;
        if (type && !headers.has("content-type")) {
          headers.append("content-type", type);
        }
      }
      if (config.proxyAgent) {
        const proxyURL = new URL(config.proxyAgent);
        const factory = this._proxies[proxyURL.protocol.slice(0, -1)];
        if (!factory) throw new Error(`Cannot resolve proxy agent ${proxyURL}`);
        init.dispatcher = factory(proxyURL);
      }
      const raw = await this.ctx.waterfall("http/fetch", url, init, config, () => {
        return this.undici.fetch(url, init);
      }).catch((cause) => {
        if (_HTTP.Error.is(cause)) throw cause;
        const error = new _HTTP.Error(`fetch ${url} failed`);
        error.cause = cause;
        throw error;
      });
      const response = {
        data: null,
        url: raw.url,
        status: raw.status,
        statusText: raw.statusText,
        headers: raw.headers
      };
      const validateStatus2 = config.validateStatus ?? (() => true);
      if (!validateStatus2(raw.status)) {
        const error = new _HTTP.Error(raw.statusText);
        error.response = response;
        try {
          response.data = await this.defaultDecoder(raw);
        } catch {
        }
        throw error;
      }
      if (config.responseType) {
        let decoder;
        if (typeof config.responseType === "function") {
          decoder = config.responseType;
        } else {
          decoder = this._decoders[config.responseType];
          if (!decoder) {
            throw new TypeError(`Unknown responseType: ${config.responseType}`);
          }
        }
        response.data = await decoder(raw);
      } else {
        response.data = await this.defaultDecoder(raw);
      }
      return response;
    } finally {
      dispose();
    }
  }
  async head(url, config) {
    const response = await this(url, { method: "HEAD", validateStatus, ...config });
    return response.headers;
  }
  ws(url, _config) {
    const config = this.resolveConfig(_config);
    url = this.resolveURL(url, config, true);
    const headers = new Headers(config.headers);
    const init = {
      headers
    };
    if (config.proxyAgent) {
      const proxyURL = new URL(config.proxyAgent);
      const factory = this._proxies[proxyURL.protocol.slice(0, -1)];
      if (!factory) throw new Error(`Cannot resolve proxy agent ${proxyURL}`);
      init.dispatcher = factory(proxyURL);
    }
    this.ctx.emit(this, "http/websocket-init", url, init, config);
    const socket = new this.undici.WebSocket(url, init);
    const dispose = this.ctx.on("dispose", () => {
      socket.close(1e3, "context disposed");
    });
    socket.addEventListener("close", () => {
      dispose();
    });
    return socket;
  }
};
var index_default = HTTP;
export {
  HTTP,
  index_default as default
};
